/* ***************************************************************** *
 *  GetPcx.C - Contains functions GetPCX          *
 *  Copyright (c) 1987-1992   Metagraphics Software Corporation      * 
 * ***************************************************************** */

#ifndef  cMET       /* If compiled separately, not "included" */
#include <stdlib.h>
#include <stddef.h>
#include <stdio.h>
#include <string.h>
#include <conio.h>
#include <dos.h>
#include <ctype.h>
#include "metawndo.h"      /* master MetaWINDOW include file */
#include "pcxhead.h"  
#endif /* separate */

#define  PACKED   0xC0  /* two hi bits of byte */ 
#define  MASK     0x3F  /* lower six bits */

/*
This function reads the PCX header record and palette information
for a specified PCX file.
*/



/*
This function copies the picture image in the specified PCX file to
the current port and bitmap. Note that using "sourceR" to read in
a specified rectangular area has not been implemented yet and will
result in a compiler warning.
*/

int SEQ GetPCX(void* aPcxFileImage, rect* sourceR, rect* destR )
	{
	char* pcxFileImage;
	register    Byte  unwork, unwork2;
	Byte        *unpkPtr,*wrkPtr;
	Word        x,y,z,i,lineCnt,height;
	rect        workR, R;
	grafPort    *currPort, pcxPort;
	grafMap     *currBitmap;
	imageHeader *imPtr;   
	pcxHead     pcxInfo;
	rect			newClip;

	pcxFileImage = (char*)aPcxFileImage;


	GetPort(&currPort);
	currBitmap = currPort->portMap;

	/* Read Header Record */
	memmove(&pcxInfo, pcxFileImage, sizeof(pcxInfo));
	pcxFileImage += sizeof(pcxInfo);
   
	if((pcxInfo.pcxBits != (Byte)currBitmap->pixBits) ||
		(pcxInfo.pcxPlanes != (Byte)currBitmap->pixPlanes) )
		{
		return 0;
		}
      

	lineCnt = pcxInfo.pcxRowBytes * pcxInfo.pcxPlanes;  

	if((unpkPtr = (Byte *)(malloc(lineCnt + sizeof(imageHeader)))) == NULL)
		{
		return 0;
		}

	/* Fill in header for WriteImage */
	imPtr = (imageHeader *) unpkPtr;

	/* PCX info considers Xmax to be Xmax-1 (ie 639 for 640x480) */
	imPtr->imWidth    = (pcxInfo.pcxWndo.Xmax+1) - pcxInfo.pcxWndo.Xmin;
	imPtr->imHeight   = 1;
	imPtr->imAlign    = 0;
	imPtr->imRowBytes = pcxInfo.pcxRowBytes;
	imPtr->imBits     = pcxInfo.pcxBits;
	imPtr->imPlanes   = pcxInfo.pcxPlanes;


	/* check if the destination rect pointer is null */
	if(destR == NULL)
		DupRect( &(currPort->portRect), &workR );  /* use current port's rect */
	else
		DupRect( destR, &workR );

	/* must do all work in global coordinates */
	memcpy(&newClip, &currPort->portClip, sizeof (rect));
	if(currPort->portFlags & pfVirtual) /* using virtual coordinates? */
		{
		Vir2GblRect( &workR );        /* translate rect to read file */
		Vir2GblRect( &newClip );
		}
	else /* in case we are using local coordinates */
		{
		Lcl2GblRect( &workR );
		Lcl2GblRect( &newClip);
		}
	// bbs600 InitPort( &pcxPort );     /* use a fresh port with everything in global */
	// bbs600 PortBitmap( currBitmap ); /* attach the current bitmap to it            */
	// bbs600 PortSize( currBitmap->pixWidth, currBitmap->pixHeight );
	// bbs600 ClipRect( &newClip );

	/* calculate the rectangle to write */
	R.Xmin = workR.Xmin;
	R.Xmax = imPtr->imWidth + workR.Xmin;
	height = pcxInfo.pcxWndo.Ymax - pcxInfo.pcxWndo.Ymin;

/*
** Begin. Unpack the byte stream into an intermediate buffer.
** This is a much simpler method when dealing with multiplanar source
** files, as it is packed into a continuous stream.
*/

	/* one  'y' scan line at a time. */ 
	for( y = 0; y <= height; y++) 
		{ 
		/* Adjust buffer pointer for data stream. */
		wrkPtr = unpkPtr + sizeof(imageHeader);

		/* set rectangle up */
		R.Ymin = R.Ymax = workR.Ymin + y; 
		R.Ymax = R.Ymin + 1;  /* write image one scan line */

		/* build the raster lines */
		for(x = 0;  x < lineCnt;   ) 
			{
			/* get the byte in question */
			unwork = *pcxFileImage;
			pcxFileImage++;

			if((unwork & PACKED) == PACKED) 
				{
	            /* calculate byte value */
		        z = unwork & MASK;

	            unwork2 = *pcxFileImage;
				pcxFileImage++;


		        /* unpack info to buffer */
	            for( i = 0; i < z; i++ ) *wrkPtr++ = unwork2;

				x += z; 
				}
			else
				{ /* not packed */
				*wrkPtr++ = unwork;         
				x++;
				}
			}  /* next picture dot in x direction */

		/* write to current ports bitmap */
		WriteImage( &R,unpkPtr );
		}  /* next multiplane raster line */

	/* DONE ! */
	SetPort(currPort);   /* restore original port */
	free( unpkPtr ); /* unpacked line */
	return 1;
	}
