/* ***************************************************************** *
 *  READPCX.C - Contains functions ReadPCXInfo and ReadPCX           *
 *  Copyright (c) 1987-1992   Metagraphics Software Corporation      * 
 * ***************************************************************** */

#ifndef  cMET       /* If compiled separately, not "included" */
#include <stdlib.h>
#include <stddef.h>
#include <stdio.h>
#include <string.h>
#include <conio.h>
#include <dos.h>
#include <ctype.h>
#include "metawndo.h"      /* master MetaWINDOW include file */
#include "pcxhead.h"  
#endif /* separate */

#define  PACKED   0xC0  /* two hi bits of byte */ 
#define  MASK     0x3F  /* lower six bits */


/*
This function reads the PCX header record and palette information
for a specified PCX file.
*/

int SEQ ReadPCXInfo( char *fname, pcxHead *pcxHdr, palData RGBColors[] )
{
   int       i, j;
   long int  fpos;
   FILE     *datfile;


   /* Open source file */
   if( ( datfile = fopen( fname,"r+b") ) == NULL ) /* binary mode read */
      return 0;

   /*  Read Header */
   fread( pcxHdr, sizeof(pcxHead), 1, datfile);

   /* Validate PCX file by checking */
	/* manufacturers ID (Zsoft-10) and PCX encoding (1) */
   if( (pcxHdr->pcxManf != 10) || (pcxHdr->pcxEncode != 1) )
	{
		fclose( datfile );
	   return 0;
   }

   /* convert PCX digital palette data into MetaWINDOW compatible format */
   for( i = 0; i < 16; i++ )
   {
      j = i*3;
      RGBColors[i].palRed   = (int) pcxHdr->pcxColor[j];
      RGBColors[i].palGreen = (int) pcxHdr->pcxColor[j+1];
      RGBColors[i].palBlue  = (int) pcxHdr->pcxColor[j+2];

      /* scale from 8 to 16 bit values */
      RGBColors[i].palRed   *= 256;
      RGBColors[i].palGreen *= 256;
      RGBColors[i].palBlue  *= 256;
   }

	/* check for optional 256 color palette */
   fseek( datfile, 0L, SEEK_END );   /* seek to 769 bytes from end of file */
   fpos = ftell( datfile );
   fseek( datfile, fpos - 769L, SEEK_SET ); 

   /* should be a 12 there */
   i = getc( datfile );
   if( i != 12 )
	{
      fclose( datfile );
      return 1;           /* analog palette data is optional */
   }
   
   /* convert PCX analog palette data into WritePalette compatible format */
   for( i=0; i<256; i++ )
   {
      RGBColors[i].palRed   = getc(datfile);
      RGBColors[i].palGreen = getc(datfile);
      RGBColors[i].palBlue  = getc(datfile);

      /* scale from 8 to 16 bit values */
      RGBColors[i].palRed   *= 256;
      RGBColors[i].palGreen *= 256;
      RGBColors[i].palBlue  *= 256;
	}

   fclose( datfile );
   return 1;
}


/*
This function copies the picture image in the specified PCX file to
the current port and bitmap. Note that using "sourceR" to read in
a specified rectangular area has not been implemented yet and will
result in a compiler warning.
*/

int SEQ ReadPCX( char *fname, rect *sourceR, rect *destR )
{
   register    Byte  unwork, unwork2;
   Byte        *unpkPtr,*wrkPtr;
   Word        x,y,z,i,lineCnt,height;
   FILE        *datfile;
   rect        workR, R;
   grafPort    *currPort, pcxPort;
   grafMap     *currBitmap;
   imageHeader *imPtr;   
	pcxHead     pcxInfo;
	rect			newClip;


   GetPort(&currPort);
   currBitmap = currPort->portMap;

   /* Open source file */
   if( ( datfile = fopen( fname,"r+b") ) == NULL ) /* binary mode read */
      return 0;

   /* Read Header Record */
   fread( &pcxInfo, sizeof(pcxInfo), 1, datfile);
   
   if( (pcxInfo.pcxBits != (Byte)currBitmap->pixBits) ||
                             (pcxInfo.pcxPlanes != (Byte)currBitmap->pixPlanes) )
   {
      fclose( datfile );
      return 0;
   }
      

   lineCnt = pcxInfo.pcxRowBytes * pcxInfo.pcxPlanes;  

   if( ( unpkPtr = (Byte *)(malloc( lineCnt + sizeof(imageHeader) ) ) ) == NULL  )
   {
      fclose( datfile );
      return 0;
   }

   /* Fill in header for WriteImage */
   imPtr = (imageHeader *) unpkPtr;

   /* PCX info considers Xmax to be Xmax-1 (ie 639 for 640x480) */
   imPtr->imWidth    = (pcxInfo.pcxWndo.Xmax+1) - pcxInfo.pcxWndo.Xmin;
	imPtr->imHeight   = 1;
   imPtr->imAlign    = 0;
   imPtr->imRowBytes = pcxInfo.pcxRowBytes;
   imPtr->imBits     = pcxInfo.pcxBits;
   imPtr->imPlanes   = pcxInfo.pcxPlanes;


   /* check if the destination rect pointer is null */
   if ( destR == NULL )
      DupRect( &(currPort->portRect), &workR );  /* use current port's rect */
   else
      DupRect( destR, &workR );

	/* must do all work in global coordinates */
	memcpy(&newClip, &currPort->portClip, sizeof (rect));
	if ( currPort->portFlags & pfVirtual ) /* using virtual coordinates? */
	{
		Vir2GblRect( &workR );        /* translate rect to read file */
		Vir2GblRect( &newClip );
	}
	else /* in case we are using local coordinates */
	{
		Lcl2GblRect( &workR );
		Lcl2GblRect( &newClip);
	}
	InitPort( &pcxPort );     /* use a fresh port with everything in global */
	PortBitmap( currBitmap ); /* attach the current bitmap to it            */
	PortSize( currBitmap->pixWidth, currBitmap->pixHeight );
	ClipRect( &newClip );

   /* calculate the rectangle to write */
   R.Xmin = workR.Xmin;
   R.Xmax = imPtr->imWidth + workR.Xmin;
   height = pcxInfo.pcxWndo.Ymax - pcxInfo.pcxWndo.Ymin;
/*
** Begin. Unpack the byte stream into an intermediate buffer.
** This is a much simpler method when dealing with multiplanar source
** files, as it is packed into a continuous stream.
*/

   /* one  'y' scan line at a time. */ 
   for( y = 0; y <= height; y++) { 

      /* Adjust buffer pointer for data stream. */
      wrkPtr = unpkPtr + sizeof(imageHeader);

      /* set rectangle up */
      R.Ymin = R.Ymax = workR.Ymin + y; 
      R.Ymax = R.Ymin + 1;  /* write image one scan line */


      /* build the raster lines */
      for( x = 0;  x < lineCnt;   ) {

         /* get the byte in question */
			unwork= (Byte) getc(datfile);

         if( (unwork & PACKED) == PACKED ) {

            /* calculate byte value */
            z = unwork & MASK;

            unwork2 = (Byte) getc(datfile);

            /* unpack info to buffer */
            for( i = 0; i < z; i++ ) 
               *wrkPtr++ = unwork2;

            x += z; 
         }
         else { /* not packed */
            *wrkPtr++ = unwork;         
            x++;
			}

      }  /* next picture dot in x direction */

      /* write to current ports bitmap */
      WriteImage( &R,unpkPtr );

   }  /* next multiplane raster line */


   /* DONE ! */
   SetPort(currPort);   /* restore original port */
   free( unpkPtr ); /* unpacked line */
   fclose( datfile );
   return 1;
}
