/* ***************************************************************** *
 *  WRITEPCX.C - Contains function WritePCX                          *
 *  Copyright (c) 1987-1992   Metagraphics Software Corporation      * 
 * ***************************************************************** */

#ifndef  cMET       /* If compiled separately, not "included" */
#include <stdlib.h>
#include <stddef.h>
#include <stdio.h>
#include <string.h>
#include <conio.h>
#include <dos.h>
#include <ctype.h>
#include "metawndo.h"      /* master MetaWINDOW include file */
#include "pcxhead.h"  
#endif /* separate */

#define  PACK 0xC0      /* two hi bits of byte */ 

int writ_byts( short wb_count, register Byte wb_last, FILE *wb_datfile );


/*
This function copies the pixel image from the currnet port and bitmap
to the specified PCX file.
*/

int SEQ WritePCX( char *fname, rect *sourceR, palData RGBColors[] )
{
   grafPort       *currPort, pcxPort;  
   grafMap        *currBitmap;
   pcxHead        pcxInfo;
   FILE           *datfile;
   int            i, y, p, rowCnt, count, align;
   long           imBytes;
   rect           workR, R;
   Byte           next, curr, last;
   Word           tmp_curr;
   Byte           *pkPtr, *bufPtr, *wrkPtr;
   imageHeader    *imPtr;


   GetPort(&currPort);
   currBitmap = currPort->portMap;

   /* check if the source rect pointer is null */
   if ( sourceR == NULL )
      DupRect( &(currPort->portRect), &workR ); /* use current port's rect */
   else
      DupRect( sourceR, &workR );

   /* must do all work in global coordinates */
   if ( currPort->portFlags & pfVirtual ) /* using virtual coordinates? */
      Vir2GblRect( &workR );
   else
      Lcl2GblRect( &workR );    /* in case we are using local coordinates */

   InitPort( &pcxPort );   /* use a fresh port with everything in global */


/*
** Calculate one raster line of the rectangle to write
*/ 
   R.Xmin= workR.Xmin;
   R.Xmax= workR.Xmax; 
   R.Ymin= 0;
   R.Ymax= 1;

/*
** allocate an image buffer for one raster line
*/
   imBytes = ImageSize(&R);   /* get size to malloc */
   imBytes += 2;              /* extra so we can 'look ahead' */
   if( ( pkPtr = (Byte *)( malloc( (unsigned int)imBytes ) ) )  == NULL )
      return 0;

/*
** read the first raster line into an image buffer
*/
   ReadImage(&R, pkPtr);
   imPtr = (imageHeader *)pkPtr;       /* pointer to MW image header */

/*
** Fill in the PCX File Header
*/
   pcxInfo.pcxManf = 10;         
   pcxInfo.pcxVers = 5;         
   pcxInfo.pcxEncode = 1;     
   pcxInfo.pcxBits = (Byte)imPtr->imBits;
   pcxInfo.pcxPlanes = (Byte)imPtr->imPlanes;    
   pcxInfo.pcxRowBytes = imPtr->imRowBytes;
   pcxInfo.pcxWndo = workR;         
   pcxInfo.pcxResX = currBitmap->pixResX;         
   pcxInfo.pcxResY = currBitmap->pixResY;          
   pcxInfo.pcxRsvd = 0;         

/*
** Fill in the digital color palette
*/

   for( i=0; i<16; i++ )
   {
      /* store the values into header */
      /* PCX specifies 8 bit color values */
      p = i*3;
      pcxInfo.pcxColor[p+0]= (Byte)(RGBColors[i].palRed/256); 
      pcxInfo.pcxColor[p+1]= (Byte)(RGBColors[i].palGreen/256); 
      pcxInfo.pcxColor[p+2]= (Byte)(RGBColors[i].palBlue/256); 
   }

/*
** Open the file
*/
   if( (datfile=fopen( fname,"w+b")) == NULL)    /* binary mode write */
      return 0;

/*
** Write Header
*/
   if( fwrite(&pcxInfo, sizeof(pcxHead), 1, datfile) == 0)  {
      fclose(datfile);
      return 0;
   }

/*
** Packing and Writing Screen to PCX file 
*/
   /* one  'y' scan line at a time. */ 
   for( y = 0; y <= workR.Ymax-workR.Ymin; y++)
   {
      /* set-up next raster line */
      R.Ymin= workR.Ymin+y;
      R.Ymax= R.Ymin+1;

      ReadImage(&R, pkPtr);

      /* Adjust buffer pointer for data stream. */
      bufPtr = pkPtr + sizeof(imageHeader);

      /* shift all image data left by imAlign pixels */
      align = imPtr->imAlign * imPtr->imBits;
      count = imPtr->imRowBytes * imPtr->imPlanes;
      wrkPtr = bufPtr;

      /* if shifting less than a byte, use this loop */
      if( align > 0 && align < 8 ) {

         align = 8 - align; /* convert left shift to right shift */

         for( i = 0; i < count; i++) {
            curr = *wrkPtr;                            
            next = *(wrkPtr+1);                        
            tmp_curr = (curr << 8) + next;  /* make word */       
            *wrkPtr++ = (Byte) ( tmp_curr >> align );
         }

      }

      /* if shifting more than a byte, use this loop */
      if( align >= 8 ) {

         align -= 8;        /* implied byte shift included */
         align = 8 - align; /* convert left shift to right shift */

         for( i = 0; i < count; i++ ) {
            curr = *(wrkPtr+1);                            
            next = *(wrkPtr+2);           
            tmp_curr = (curr << 8) + next; /* make word */         
            *wrkPtr++ = (Byte) ( tmp_curr >> align );
         }

      }

         
      /* do all planes */
      for ( p = 0; p < imPtr->imPlanes; p++ ) {

         wrkPtr = bufPtr;

         count = 1;
         last = *wrkPtr;      /* get the byte in question */
         
         /* write out all bytes in row (raster) */
         for( rowCnt = 1; rowCnt < imPtr->imRowBytes; rowCnt++) {

            curr = *(++wrkPtr);  /* get the next byte */
         
            if (curr == last) {

               count++;

               if (count == 63) {   /* count can only be six bits long (63) */
                  if ( writ_byts( count, last, datfile) )
                     return 0;
                  count = 0;   
               }

            }
            else  { 
               if (count) {
                  if ( writ_byts( count, last, datfile) )
                     return 0;
               } 
               last = curr;

               count = 1;

            }

         }    /* for all bytes in row (raster) */
   
         if (count) {
            if ( writ_byts( count, last, datfile) )
               return 0;
         }

         bufPtr += imPtr->imRowBytes;  /* point to beginning of next plane */
              
      }    /* for loop for all planes */

   }    /* for loop all of raster lines */

   SetPort(currPort);   /* restore original port */
   free( pkPtr );       /* free packing buffer */


   /* if 256 color modes, write out the analog palette */
   if( QueryColors() != 255 || pcxInfo.pcxVers < 5  ) {
      fclose( datfile );
      return 1; 
   }

   /* write 'magic number', signature byte */
   putc( 12, datfile );

   for( i = 0; i < 256; i++ ) {

      /* append scaled values to file */
      /* PCX specifies 8 bit color values */
      putc( RGBColors[i].palRed/256,   datfile );
      putc( RGBColors[i].palGreen/256, datfile );
      putc( RGBColors[i].palBlue/256,  datfile );

   }

/*
** DONE !
*/

   fclose( datfile );
   return 1;

}


int writ_byts( short wb_count, register Byte wb_last, FILE *wb_datfile )
{
   if(wb_count) {

      if( (wb_count == 1) && (PACK != (PACK&wb_last)) ) {
         if( EOF == putc( (int)wb_last, wb_datfile))
            return 1;  /* disk write error */
      }

      else  { /* pack bytes */

         if( EOF == putc( (int)PACK | wb_count, wb_datfile))
            return 1;  /* disk write error */
         if( EOF == putc( (int)wb_last, wb_datfile))    
            return 1;  /* disk write error */
      }

   }

   return 0;

} /* end of writ_byts */

   
